!+
! Module: dpssflag_mod
!
! Authors:  Maite Ceballos - IFCA
!          
! Description: Flags sources in EPIC combined source list according 
! to several criteria
!     
!
! Comments: 
! 
! $Id: dpssflag_mod.f90,v 4.0 2005/12/13 13:45:39 MCeballos Exp $
!- 
module dpssflag_mod

  use dal
  use cal
  use coordinate
  use wcs_aux
  use param  
  use errorHandling
  use constants
  use dal_aux


  implicit none 
  
  private 

  integer(kind=int32),dimension(:),pointer :: id,idinst,idband,rawx,rawy
  integer(kind=int32),dimension(:),pointer :: ccdnr
  real(kind=single),dimension(:),pointer :: maskFrac,bright,distnn,ext
  real(kind=single),dimension(:),pointer :: detml,scts,badpix !,extml
  real(kind=double),dimension(:),pointer :: ra,dec

  integer(kind=int32):: nrows
  real(kind=double):: expStartTimeStamp
  real(kind=single) :: mfraclimit,brfactor,extfactor,extlimit,badfactor
  real(kind=single) :: brexfraction,brlimit,maxextrad
  real(kind=single) :: detmlfraction,degimpix,minrad,maxrad,fixext
  character(len=256) ::srclist,srclistSetName,srclistTabName
  character(len=10) :: brcolumn,badcolumn
  character(len=20) :: pnsubmode,mos1submode
  character(len=256) :: dateTimeStr
  character(len=4) :: op1
  character(len=3) :: op2
  logical :: pnflag9on

  type(DataSetT):: inSet
  type(TableT)::  inTab
  type(ColumnT):: coID,coRA,coDEC,coMASKFRAC,coBRIGHT,coDISTNN,coFLAG
  type(ColumnT)::coEXT,coIDINST,coIDBAND,coDETML,coSCTS,coBADPIX!,coEXTML
  type(ColumnT):: coRAWX,coRAWY,coCCDNR
  type(SpacecraftAttitudeType) :: scAttitude

  public dpssflag

contains 
  
  !---------------------------------------------------------------------
  subroutine dpssflag
  !---------------------------------------------------------------------
    implicit none
        
!   read input parameters from param file

    srclist=stringParameter("srclisttable")  

    mfraclimit=realParameter("mfraclimit")      !Flag1

    brcolumn=stringParameter("brcolumn")        !Flag2
    op1=stringParameter("op1")       
    op2=stringParameter("op2")       
    if(op2 == "add" .OR. op2 == "mul")then
       brfactor=realParameter("brfactor")
    endif
    minrad=realParameter("minrad")
    maxrad=realParameter("maxrad")

    extfactor=realParameter("extfactor")        !Flag3
    extlimit=realParameter("extlimit")
    maxextrad=realParameter("maxextrad")

    brlimit=realParameter("brlimit")            !Flag4

    brexfraction=realParameter("brexfraction")  !Flag5
    fixext=realParameter("fixext")              !Flag5               

    detmlfraction=realParameter("detmlfraction")!Flag6

    badcolumn=stringParameter("badcolumn")      !Flag8
    badfactor=realParameter("badfactor")

    pnflag9on=booleanParameter("pnclosetobadon")   !Flag9

    call splitSetTabName(srclist,srclistSetName,srclistTabName)
    
    inSet=dataSet(srclistSetName,MODIFY)
    inTab=table(inSet,srclistTabName,MODIFY)
    nrows=numberOfRows(inTab)
    pnsubmode=stringAttribute(inSet,"SUBM_PN")
    mos1submode=stringAttribute(inSet,"SUBM_M1")
    degimpix=real32Attribute(inSet,"TCDELT1")
    dateTimeStr=stringAttribute(inSet,"DATE-OBS")
    scAttitude%ra=real32Attribute(inSet,"RA_PNT")
    scAttitude%dec=real32Attribute(inSet,"DEC_PNT")
    scAttitude%apos=real32Attribute(inSet,"PA_PNT")
    expStartTimeStamp=toTimeTag(dateTimeStr)
    scAttitude%ra=scAttitude%ra*DegToRad
    scAttitude%dec=scAttitude%dec*DegToRad
    scAttitude%apos=scAttitude%apos*DegToRad


!   read columns in source list input table
    coID=column(inTab,"ML_ID_SRC",READ)
    coEXT=column(inTab,"EXT",READ)
    coRA=column(inTab,"RA",READ)
    coDEC=column(inTab,"DEC",READ)
    coMASKFRAC=column(inTab,"MASKFRAC",READ)
    coBRIGHT=column(inTab,trim(brcolumn),READ)
    coIDINST=column(inTab,"ID_INST",READ)
    coIDBAND=column(inTab,"ID_BAND",READ)
    coDETML=column(inTab,"DET_ML",READ)
!    coEXTML=column(inTab,"EXT_ML",READ)
    coSCTS=column(inTab,"SCTS",READ)
    coDISTNN=column(inTab,"DIST_NN",READ)
    coFLAG=column(inTab,"FLAG",MODIFY)
    coBADPIX=column(inTab,trim(badcolumn),READ)
    coRAWX=column(inTab,"RAWX",READ)
    coRAWY=column(inTab,"RAWY",READ)
    coCCDNR=column(inTab,"CCDNR",READ)

!   read data in columns
    id=>int32Data(coID)
    ra=>real64Data(coRA)
    dec=>real64Data(coDEC)
    maskFrac=>real32Data(coMASKFRAC)
    bright=>real32Data(coBRIGHT)
    idinst=>int32Data(coIDINST)
    idband=>int32Data(coIDBAND)
    detml=>real32Data(coDETML)
!    extml=>real32Data(coEXTML)
    distnn=>real32Data(coDISTNN)
    ext=>real32Data(coEXT)
    scts=>real32Data(coSCTS)
    badpix=>real32Data(coBADPIX)
    rawx=>int32Data(coRAWX)
    rawy=>int32Data(coRAWY)
    ccdnr=>int32Data(coCCDNR)

    !initialize flag sequence to F and set 1st Flag: low MASKFRAC value
    call setFlag1()


    ! set 2nd flag: source in 'circle_of_influence' of a point source
    ! set 4th flag: ext source in 'circle_of_influence' of a point source
    !               with more than "brlimit" counts
    call setFlag24()

    ! set 3rd flag: source in 'circle_of_extended emission' of an ext source
    ! set 5th flag: ext source in 'circle_of_extended emission' of an ext 
    !               brighter source (flagged <brexfraction * brighter)
    call setFlag35()

    ! set 6th flag: extended and det_ml(iband)> detmlfraction * detmlsum(1:15)
    call setFlag6();

    ! set 7th flag: likely spurious extended: Flag4 .or. Flag5 .or. Flag6 
    call setFlag7();

    ! set 8th flag: on bad PN column if INSTRUME=PN (idinst=1)
    ! set 9th flag: close bad PN column if INSTRUME=PN (idinst=1)
    call setFlag89();
    
    call release(inSet)
    
  end subroutine dpssflag

!--------------------------------------------------------------------------
!
!
!     SUBROUTINE SETFLAG1()
!
!     initialize flag sequence to F and set 1st Flag: low MASKFRAC value
!--------------------------------------------------------------------------

  subroutine setFlag1()
    ! it uses variables: inTab,coFLAG,maskFrac,mfraclimit
    implicit none
    integer(kind=int32):: i
    character(len=12) :: newflag
    character(len=1), dimension(:), allocatable :: flag1
    character(len=12) :: flagStr

    allocate(flag1(0:nrows-1))
    flag1="F"

    flagStr="FFFFFFFFF"
    do i=0,nrows-1
       call setStringCell(coFLAG,i,flagStr)
       if(.not.(scts(i)>=0.) .AND. .not.(scts(i)<0.))cycle  !NULL row
       if(maskFrac(i) <=mfraclimit)then
          newflag='TFFFFFFFF'
          call setStringCell(coFLAG,i,newflag)
          flag1(id(i))="T"
       endif
    enddo

    !set also flag 1 for row with idinst = 0 (summary row)
    do i=0,nrows-1
       if(idinst(i)==0 .AND. flag1(id(i))=="T")then
          flagStr=stringCell(coFLAG,i)
          flagStr(1:1)="T"
          call setStringCell(coFLAG,i,flagStr)
       endif
    enddo

  end subroutine setFlag1


!--------------------------------------------------------------------------
!
!
!     SUBROUTINE SETFLAG24()
!
!      Set 2nd Flag: source close to another bright point source
!      and 4th Flag: ext source close to a bright (sum_row_cts>brlimit) 
!                    point source
!--------------------------------------------------------------------------

  subroutine setFlag24()

    !it uses: inTab,id,bright,brfactor,distnn,coFlag
    integer(kind=int32):: i,j,jsumm
    real(kind=single):: cOfInf,cOfInf1
    real(kind=double) :: dist
    character(len=12) :: flagStr,flagStrP
    logical :: summrow = .FALSE.

    if(idinst(0)==0 .AND. idband(0)==0) summrow = .TRUE.

    do i=0,nrows-1
       ! foreach 'point' source in list: calculate 'circle of influence' 

       if(i>0)then !use summary row (idinst=idband=0) or idband=0 if only 1 ins
          if(id(i)==id(i-1))cycle
       endif
       if(ext(i) >= extlimit)cycle !only for point sources
       if(op1 == "sqrt")then
          cOfInf1=sqrt(bright(i))
       elseif(op1 == "none")then
          cOfInf1=bright(i)
       else
          call fatal("UknOpFlag2",'Unknown operation for Flag2 expression');
       endif

       if(op2 == "add")then
          cOfInf=cOfInf1 + brfactor
       elseif(op2 == "mul")then
          cOfInf=cOfInf1 * brfactor
       elseif(op1 == "non")then
          cOfInf=cOfInf1
       else
          call fatal("UknOpFlag2",'Uknown operation for Flag2 expression');
       endif

!       cOfInf=bright(i)*brfactor  !arcsec
       if(cOfInf < minrad)cOfInf=minrad
       if(cOfInf > maxrad)cOfInf=maxrad
       
       if(cOfInf>=distnn(i))then  !if there are neighbours close to source

          do j=0,nrows-1  !for each posible neighbour

             ! find which id the summary row for later flag copying
             if(idinst(j)==0 .AND. idband(j)==0 .AND. summrow) then
                jsumm=j
             elseif (.NOT. summrow .AND. idband(j)==0)then
                jsumm=j
             endif
             if(.not.(scts(j)>=0.) .AND. .not.(scts(j)<0.))cycle  !NULL row
             if (id(i) /= id(j))then    !skip all rows for same source
                flagStr=stringCell(coFLAG,j)
                if(j>0)then
                   !if same neighb., just copy flags 2&4 from summary row
                   if(id(j)==id(j-1))then 
                      flagStrP=stringCell(coFLAG,jsumm)
                      flagStr(2:2)=flagStrP(2:2)
                      flagStr(4:4)=flagStrP(4:4)
                      call setStringCell(coFLAG,j,flagStr)
                      cycle
                   endif
                endif
                !calculate distance 
                dist=distance(ra(i),dec(i),ra(j),dec(j))
                !if neighbour in circle of influence --> flag neighbour 2 (& 4)
                if(dist < cOfInf)then
                   flagStr(2:2)='T'
                   if(ext(j)>=extlimit .AND. scts(i)>brlimit) &
                        flagStr(4:4)='T'
                   call setStringCell(coFLAG,j,flagStr)
                endif
             endif
          enddo !neighbours loop
       endif
    enddo !sources loop
  end subroutine setFlag24

!--------------------------------------------------------------------------
!
!
!     SUBROUTINE SETFLAG35()
!
!      Set 3rd Flag: source in area of extended emission of another source
!      Set 5th Flag: ext source in area (fixed radius) of extended emission  
!                    of another extended and brighter source
!--------------------------------------------------------------------------

  subroutine setFlag35()

    !it uses: inTab,id,ext,extfactor,distnn,coFlag
    integer(kind=int32):: i,j,jsumm
    real(kind=single):: cOfExt,cOfExtF5
    real(kind=double) :: dist
    character(len=12) :: flagStr,flagStrP
    logical :: summrow = .FALSE.

    if(idinst(0)==0 .AND. idband(0)==0) summrow = .TRUE.

    do i=0,nrows-1
       if(i>0)then !use summary row (idinst=idband=0) or idband=0 if only 1 ins
          if(id(i)==id(i-1))cycle
       endif
       if(ext(i)<extlimit)cycle   !if source 'not extended' -> next source
       ! foreach source in list: calculate 'circle of extension' based 
       ! on source EXT info
       cOfExt=ext(i)*extfactor  !pixel
       cOfExt=cOfExt*ABS(degimpix)*3600.! arcs
       cOfExtF5=fixext !arcs
       if(cOfExt>maxextrad)cOfExt=maxextrad

       !if there are neighbours close to source:
       if(cOfExt>=distnn(i) .OR. cOfExtF5>=distnn(i))then  
          do j=0,nrows-1  !for each posible neighbour
             if(idinst(j)==0 .AND. idband(j)==0 .AND. summrow) then
                jsumm=j
             elseif (.NOT. summrow .AND. idband(j)==0)then
                jsumm=j
             endif
             if(.not.(scts(j)>=0.) .AND. .not.(scts(j)<0.))cycle  !NULL row
             if (id(i) /= id(j))then    !skip all rows for same source
                flagStr=stringCell(coFLAG,j)
                if(j>0)then
                   !if same neighb., just copy flags 3&5 from summary row
                   if(id(j)==id(j-1))then 
                      flagStrP=stringCell(coFLAG,jsumm)
                      flagStr(3:3)=flagStrP(3:3)
                      flagStr(5:5)=flagStrP(5:5)
                      call setStringCell(coFLAG,j,flagStr)
                      cycle
                   endif
                endif
                !calculate distance
                dist=distance(ra(i),dec(i),ra(j),dec(j))
                !if neighbour in circle of ext emission --> flag 3 
                if(dist < cOfExt)then
                   flagStr(3:3)='T'
                   call setStringCell(coFLAG,j,flagStr)
                endif
                !if neighbour in circle of fixed ext emission & fainter
                ! than triggering source --> flag 5 
                if(dist < cOfExtF5 .AND. ext(j)>=extlimit .AND. &
                     bright(j)<brexfraction*bright(i))then
                   flagStr(5:5)='T'
                   call setStringCell(coFLAG,j,flagStr)
                endif
             endif
          enddo !neighbours loop
       endif
    enddo !sources loop
  end subroutine setFlag35


!--------------------------------------------------------------------------
!
!
!     SUBROUTINE SETFLAG6()
!
!      Set 6th Flag: ext source has det_ml(iband) > detmlfraction * det_ml_sum 
!                    
!--------------------------------------------------------------------------

  subroutine setFlag6()

    !it uses: inTab,id,ext,extfactor,distnn,coFlag
    integer(kind=int32):: i
    real(kind=single),dimension(:),allocatable:: detmltot,detmlsum
    character(len=12) :: flagStr
    character(len=1), dimension(:), allocatable :: flag6

    allocate(detmlsum(0:nrows-1))
    allocate(detmltot(0:nrows-1))
    allocate(flag6(0:nrows-1))
    detmlsum=0.
    detmltot=0.
    flag6="F"

    ! calculate det_ML_sum(15 images) and det_ml(0,0) for each source
    do i=0,nrows-1
       if(idband(i)>0 .AND. idband(i)<=5) &
            detmlsum(id(i))=detmlsum(id(i))+detml(i)
       if((idband(i)==0 .AND. idinst(i)==0) .OR. &  !use sum row if present
          (idinst(i)==0 .AND. detmltot(id(i))==0)) & !othrw idinst=0 for 1 inst
          detmltot(id(i))=detml(i)
    enddo

    ! check if source has 90% of det_ML in only one band
    do i=0,nrows-1
       if(ext(i) >= extlimit)then
          if(idband(i)>0 .AND. idband(i)<=5 .AND. &
               detml(i)>detmlfraction*detmlsum(id(i)))then
             flag6(id(i))="T"
          endif
       endif
    enddo

    ! set flag6 value
    do i=0,nrows-1
       if(flag6(id(i)) .EQ. "T")then
          flagStr=stringCell(coFLAG,i)
          flagStr(6:6)="T"
          call setStringCell(coFLAG,i,flagStr)
       endif
    enddo
    deallocate(detmlsum,detmltot)
  end subroutine setFlag6

!--------------------------------------------------------------------------
!
!
!     SUBROUTINE SETFLAG7()
!
!     likely spurious extended: Flag4 .or. Flag5 .or. Flag6
!--------------------------------------------------------------------------

  subroutine setFlag7()
    implicit none
    integer(kind=int32):: i
    character(len=12) :: flagStr

    do i=0,nrows-1
       flagStr=stringCell(coFLAG,i)
       if(flagStr(4:4) .EQ. "T" .OR. flagStr(5:5) .EQ. "T" .OR. &
            flagStr(6:6) .EQ. "T")then
          flagStr(7:7) = "T"
          call setStringCell(coFLAG,i,flagStr)
       endif
    enddo

  end subroutine setFlag7



!--------------------------------------------------------------------------
!
!
!     SUBROUTINE SETFLAG89()
!
!    flag sources with distance_to_PN_band1_bad_columns <= badMaxDistRawpix
!    flag sources with distance_to_M1_CCD4_band123_bad_area <= badMaxDistRawpix
!--------------------------------------------------------------------------

  subroutine setFlag89()
    ! it uses variables: inTab,coFLAG,idinst,badpix,badfactor,pnsubmode,nrows
    implicit none
    integer(kind=int16):: ix,iy,isTherePN,isThereSummRow,rawxBA,isAdj
    integer(kind=int32):: i,nbadcolumns,yminBA,ymaxBA,ir
    integer(kind=int32):: yminBC,ymaxBC,xminBC,xmaxBC
    integer(kind=int32):: ccdBC,ccdBA,isumm
    integer(kind=int32),dimension(:,:), allocatable::bc
    real(kind=single):: badMaxDistArcs
    real(kind=double) :: tanThetaArcs,phi,dist,raBA,decBA
    real(kind=double),dimension(:,:),allocatable::raBC,decBC
    character(len=12) :: flagStr,adjStr,ccdStr
    character(len=1), dimension(:), allocatable :: flag8,flag9
    character(len=15), dimension(7) :: ccdAdjM1
    character(len=15), dimension(12) :: ccdAdjPN

    !from ssclib-3.27, to account for unresolved problem in attcalc:
    applyAttcalcFudge = .true. 

    nbadcolumns=48
    allocate(bc(0:nbadcolumns-1,0:3))
    allocate(flag8(0:nrows-1))
    allocate(flag9(0:nrows-1))
    bc=0

    ! M1 Bands 1,2,3 hot pixels CCD =1
    ccdBC=1; ccdAdjM1(1)="*6*7*"
    if(mos1submode == "PrimeFullWindow" .OR. &
         mos1submode == "PrimePartialW6")then
       xminBC=4; xmaxBC=33; yminBC=1;ymaxBC=10
    else if(mos1submode == "PrimePartialW3" .OR. &
         mos1submode == "PrimePartialW5")then
       xminBC=150; xmaxBC=182; yminBC=152;ymaxBC=166
    else if(mos1submode == "PrimePartialW2" .OR. &
         mos1submode == "PrimePartialW4")then
       xminBC=250; xmaxBC=250; yminBC=252;ymaxBC=252
    else
       mos1submode="NOFLAG"
    endif
    if(mos1submode /= "NOFLAG")then
       allocate(raBC(xminBC:xmaxBC,yminBC:ymaxBC))
       allocate(decBC(xminBC:xmaxBC,yminBC:ymaxBC))
       call CAL_setState(instrument=4,ccdChipId=ccdBC,&
            ccdModeId=3,ccdNodeId=0) !M1
       do ix=xminBC,xmaxBC
          do iy=yminBC,ymaxBC
             !transform RAWX/RAWY of hot pixel into RA/DEC
             call rawToTel(ix,iy,tanThetaArcs,phi)
             call telToRaDec(tanThetaArcs,phi,scAttitude,&
                  expStartTimeStamp,raBC(ix,iy),decBC(ix,iy))
             raBC(ix,iy)=raBC(ix,iy)*RadToDeg
             decBC(ix,iy)=decBC(ix,iy)*RadToDeg
          enddo
       enddo
    endif

    ! PN Band 1 bad columns 
    !    CCD      RAWX        YMIN        YMAX
    ccdAdjPN="******"

    bc(0,0)=1; bc(0,1)=13; bc(0,2)=0; bc(0,3)=0
    bc(1,0)=1; bc(1,1)=19; bc(1,2)=0; bc(1,3)=21
    bc(2,0)=1; bc(2,1)=45; bc(2,2)=0; bc(2,3)=0
    ccdAdjPN(1)="*2*4*7*10*11*"

    bc(3,0)=2; bc(3,1)=55; bc(3,2)=180; bc(3,3)=200
    ccdAdjPN(2)="*1*3*10*11*12*"

    bc(4,0)=3; bc(4,1)=13; bc(4,2)=0; bc(4,3)=0
    bc(5,0)=3; bc(5,1)=14; bc(5,2)=0; bc(5,3)=0
    bc(6,0)=3; bc(6,1)=28; bc(6,2)=0; bc(6,3)=0
    bc(7,0)=3; bc(7,1)=39; bc(7,2)=0; bc(7,3)=0
    ccdAdjPN(3)="*2*11*12*"

    bc(8,0)=4; bc(8,1)=0; bc(8,2)=0; bc(8,3)=0
    ccdAdjPN(4)="*1*5*7*8*10*"

    bc(9,0)=5; bc(9,1)=5; bc(9,2)=0; bc(9,3)=0
    bc(10,0)=5; bc(10,1)=11; bc(10,2)=1; bc(10,3)=200
    bc(11,0)=5; bc(11,1)=44; bc(11,2)=156; bc(11,3)=200
    ccdAdjPN(5)="*4*6*7*8*9*"

    bc(12,0)=6; bc(12,1)=21; bc(12,2)=0; bc(12,3)=0
    bc(13,0)=6; bc(13,1)=31; bc(13,2)=111; bc(13,3)=200
    ccdAdjPN(6)="*5*8*9*"

    if(pnsubmode == "PrimeFullWindowExtended") &
         bc(14,0)=7; bc(14,1)=34; bc(14,2)=50; bc(14,3)=200 !eFF
    ccdAdjPN(7)="*1*4*5*8*10*"

    bc(15,0)=8; bc(15,1)=5; bc(15,2)=0; bc(15,3)=0
    bc(16,0)=8; bc(16,1)=19; bc(16,2)=0; bc(16,3)=0
    bc(17,0)=8; bc(17,1)=27; bc(17,2)=0; bc(17,3)=0
    bc(18,0)=8; bc(18,1)=32; bc(18,2)=0; bc(18,3)=0
    ccdAdjPN(8)="*4*5*6*7*9*"

    bc(19,0)=9; bc(19,1)=11; bc(19,2)=0; bc(19,3)=0
    bc(20,0)=9; bc(20,1)=19; bc(20,2)=0; bc(20,3)=0
    bc(21,0)=9; bc(21,1)=23; bc(21,2)=0; bc(21,3)=0
    bc(22,0)=9; bc(22,1)=54; bc(22,2)=0; bc(22,3)=0
    ccdAdjPN(9)="*5*6*8*"

    bc(23,0)=10; bc(23,1)=18; bc(23,2)=75; bc(23,3)=200
    bc(24,0)=10; bc(24,1)=57; bc(24,2)=187; bc(24,3)=200
    bc(25,0)=10; bc(25,1)=61; bc(25,2)=138; bc(25,3)=200
    ccdAdjPN(10)="*1*2*4*7*11*"

    bc(26,0)=11; bc(26,1)=9; bc(26,2)=171; bc(26,3)=200
    bc(27,0)=11; bc(27,1)=12; bc(27,2)=186; bc(27,3)=200
    bc(28,0)=11; bc(28,1)=18; bc(28,2)=191; bc(28,3)=200
    bc(29,0)=11; bc(29,1)=19; bc(29,2)=171; bc(29,3)=200
    bc(30,0)=11; bc(30,1)=27; bc(30,2)=186; bc(30,3)=200
    ccdAdjPN(11)="*1*2*3*10*12*"

    if(pnsubmode == "PrimeFullWindowExtended") &
         bc(31,0)=11; bc(31,1)=27; bc(31,2)=0; bc(31,3)=24    !eFF
    bc(32,0)=11; bc(32,1)=33; bc(32,2)=186; bc(32,3)=200
    bc(33,0)=11; bc(33,1)=43; bc(33,2)=191; bc(33,3)=200
    bc(34,0)=11; bc(34,1)=44; bc(34,2)=191; bc(34,3)=200
    bc(35,0)=11; bc(35,1)=48; bc(35,2)=186; bc(35,3)=200
    bc(36,0)=11; bc(36,1)=50; bc(36,2)=191; bc(36,3)=200
    bc(37,0)=11; bc(37,1)=51; bc(37,2)=191; bc(37,3)=200
    bc(38,0)=11; bc(38,1)=62; bc(38,2)=1; bc(38,3)=200
    bc(39,0)=11; bc(39,1)=63; bc(39,2)=1; bc(39,3)=200
    bc(40,0)=11; bc(40,1)=64; bc(40,2)=1; bc(40,3)=200

    bc(41,0)=12; bc(41,1)=15; bc(41,2)=0; bc(41,3)=0
    bc(42,0)=12; bc(42,1)=16; bc(42,2)=0; bc(42,3)=0
    bc(43,0)=12; bc(43,1)=36; bc(43,2)=0; bc(43,3)=0
    bc(44,0)=12; bc(44,1)=44; bc(44,2)=169; bc(44,3)=200
    bc(45,0)=12; bc(45,1)=49; bc(45,2)=0; bc(45,3)=0
    bc(46,0)=12; bc(46,1)=53; bc(46,2)=0; bc(46,3)=0
    bc(47,0)=12; bc(47,1)=57; bc(47,2)=0; bc(47,3)=0
    ccdAdjPN(12)="*2*3*11*"


    call CAL_setState(instrument=6) !PN
    isTherePN=0
    isThereSummRow=0
    do i=0,nrows-1
       !if src not real (SCTS=NULL=NaN), cycle
       if(.not.(scts(i)>=0.) .AND. .not.(scts(i)<0.))cycle 
       if(idinst(i)==1)isTherePN=1
       if(idinst(i)==0 .AND. idband(i)==0)isThereSummRow=1
       flag8(i)="F"
       flag9(i)="F"
       if(((idinst(i).NE.1).OR.(idband(i)>1)) .AND.  &
          ((idinst(i).NE.2).OR.(idband(i) > 3))) cycle
       badMaxDistArcs=badpix(i)*badfactor*ABS(degimpix)*3600.
       !arcss        =     image pixels* (deg/impix) * (arcs/deg)
       if(idinst(i).EQ.1) then !PN bad columns
          do ir=0,nbadcolumns-1
             ccdBA=bc(ir,0)
             if(ccdBA == 0)cycle !PN in different submode
             rawxBA=bc(ir,1)
             yminBA=bc(ir,2)
             ymaxBA=bc(ir,3)
             if(yminBA == 0 .AND. ymaxBA == 0)cycle
             write(ccdStr,*) ccdnr(i)
             ccdStr=adjustl(ccdStr)
             adjStr="*"//trim(ccdStr)//"*"
             isAdj=index(ccdAdjPN(ccdBA),trim(adjStr))
!             if(ccdBA /= ccdnr(i) .AND. isAdj==0)cycle
             if(ccdBA == ccdnr(i) .AND. &
                (rawx(i) == rawxBA .AND. &
                (rawy(i)<=ymaxBA .AND. rawy(i)>=yminBA)))then 
                !srcs in same CCD than PN bcolumns
                flag8(i)="T"      ! source ON bad column
             elseif(ccdnr(i)==ccdBA .OR. isAdj/=0 .OR. &
                  rawx(i)<0 .or. rawy(i)<0 .or. ccdnr(i)<0)then 
                ! source CLOSE to bad column or  RAWX/RAWY/CCDNR are NULL
                call CAL_setState(instrument=6,ccdChipId=ccdBA) 
                BACOLS:do iy=yminBA,ymaxBA
                   !transform RAWX/RAWY of bad column pixels into RA/DEC
                   call rawToTel(rawxBA,iy,tanThetaArcs,phi)
                   call telToRaDec(tanThetaArcs,phi,scAttitude,&
                        expStartTimeStamp,raBA,decBA) ! ra/dec in rad
                   raBA=raBA*RadToDeg
                   decBA=decBA*RadToDeg
                   dist=distance(ra(i),dec(i),raBA,decBA)
                   if(dist<=badMaxDistArcs .AND. pnflag9on)then
                      flag9(i)="T"
                      EXIT BACOLS
                   endif
                enddo BACOLS
             endif
             if(flag8(i)=="T" .AND. &
                  (pnflag9on .EQV. .FALSE. .OR. flag9(i)=="T")) EXIT
          enddo !loop over bad columns
       else if(idinst(i) .EQ. 2 .AND. mos1submode /= "NOFLAG")then
          write(ccdStr,*) ccdnr(i)
          ccdStr=adjustl(ccdStr)
          adjStr="*"//trim(ccdStr)//"*"
          isAdj=index(ccdAdjM1(ccdBC),trim(adjStr))
!          if(ccdBC /= ccdnr(i) .AND. isAdj==0)cycle
          if(ccdnr(i)==ccdBC .AND.&
             (rawx(i) >= xminBC .AND. rawx(i) <= xmaxBC .AND.&
              rawy(i) <= ymaxBC .AND. rawy(i)>= yminBC)) then 
             !sources ON hot pixels area
             flag8(i)="T"      
          elseif(ccdnr(i)==ccdBC .OR. isAdj /= 0 .OR. &
               rawx(i)<0 .or. rawy(i)<0 .or. ccdnr(i)<0)then
             !sources in same or  adjacent CCCds, CLOSE to bright M1 corner
             ! OR RAWX/RAWY/CCDNR are NULL
             BAPIXS:do ix=xminBC,xmaxBC
                do iy=yminBC,ymaxBC
                   dist=distance(ra(i),dec(i),raBC(ix,iy),decBC(ix,iy))
                   if(dist<=badMaxDistArcs)then
!                      print *,"dist<badMaxDistArcs (",dist,badMaxDistArcs,"):"
!                      print *,"src is close:i,id(i)=",i,id(i) 
                      flag9(i)="T"
                      EXIT BAPIXS
                   endif
                enddo
             enddo BAPIXS
          endif
       endif
       flagStr=stringCell(coFLAG,i)
       flagStr(8:8)=flag8(i)
       flagStr(9:9)=flag9(i)
       call setStringCell(coFLAG,i,flagStr)
       !set also flags 8/9 for row with idinst = 0 (summary row)
       if(isThereSummRow==0) cycle
       isumm=-1 !write only once (not after all possible bands)
       if(idinst(i) == 1 .AND. idband(i)==1) isumm=i-2    !(2 rows upwards)
       if(idinst(i) == 2 .AND. idband(i)==3) then
          if(isTherePN==1) then
             isumm=i-10 !(10 rows upwards)
          else
             isumm=i-4 !(4 rows upwards)
          endif
       endif
       if(isumm==-1)cycle 
       !after PN band1 or M1 band3 flagging
       flagStr=stringCell(coFLAG,isumm)
       if(flag8(i) .EQ. "T") flagStr(8:8)=flag8(i)
       if(flag9(i) .EQ. "T") flagStr(9:9)=flag9(i)
       call setStringCell(coFLAG,isumm,flagStr)
    enddo


  end subroutine setFlag89




!----------------------------------------------------------------------
!
  function distance(ra1,dec1,ra2,dec2)
!
!----------------------------------------------------------------------
    implicit none
    real(kind=double),intent(IN):: ra1,dec1,ra2,dec2
    real(kind=double):: dist,distance
    
    dist=SIN(dec1*DegToRad)*SIN(dec2*DegToRad)+ &
         COS(dec1*DegToRad)*COS(dec2*DegToRad)*&
         COS((ra1-ra2)*DegToRad)
    if(ABS(dist).GT.1.) dist=dist/ABS(dist)
    distance=ACOS(dist)/DegToRad*3600.
    
  end function distance
  
end module dpssflag_mod
